from typing import Any, Optional, Type

from .ast import RqlQuery, Table
from .trio_net.net_trio import TrioConnectionPool

class RethinkDB:
    ConnectionPool: Type[TrioConnectionPool]
    row: Any
    def __init__(self) -> None: ...
    def iso8601(self, string: str, default_timezone: str = '') -> RqlQuery: ...
    def now(self) -> RqlQuery: ...
    def set_loop_type(self, library: Optional[str]) -> None: ...
    def table(self, name: str) -> Table: ...

r: RethinkDB



# Test users
# Copyright (C) 2020  Ngô Ngọc Đức Huy
# Copyright (C) 2021  Nguyễn Gia Phong
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from http import HTTPStatus as Status

from conftest import ClientFactory, parametrize
from pytest import param
from quart.testing import QuartClient


@parametrize(('username', 'status_code'),
             (param('nexistepas', Status.NOT_FOUND, id='not exist'),
              param('adaml', Status.OK, id='can have projects'),
              param('silasl', Status.OK, id='cannot have projects')))
async def test_get(username: str, status_code: int,
                   client: QuartClient) -> None:
    """Test GET user root endpoint."""
    response = await client.get(f'/u/{username}')
    assert response.status_code == status_code


@parametrize(('username', 'profile', 'status_code'),
             (('adaml', 'silasl', Status.FORBIDDEN),
              ('adaml', 'adaml', Status.OK)))
async def test_edit_get(username: str, profile: str,
                        status_code: int, user: ClientFactory) -> None:
    """Test GET user edit endpoint."""
    client = await user(username)
    response = await client.get(f'/u/{profile}/edit')
    assert response.status_code == status_code


@parametrize(('username', 'profile', 'status_code'),
             (('adaml', 'silasl', Status.FORBIDDEN),
              ('adaml', 'adaml', Status.FOUND)))
async def test_edit_post(username: str, profile: str,
                         status_code: int, user: ClientFactory) -> None:
    """Test POST user edit endpoint."""
    client = await user(username)
    info = {'name': 'Silas Salis', 'email': 'newemail@example.edu'}
    response = await client.post(f'/u/{profile}/edit', form=info)
    assert response.status_code == status_code



# Common test fixtures
# Copyright (C) 2020-2021  Nguyễn Gia Phong
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from random import choices
from string import printable
from typing import AsyncIterator, Awaitable, Callable, Optional

from pytest import fixture, mark
from quart.testing import QuartClient

from acanban import Acanban, app

ClientFactory = Callable[[Optional[str]], Awaitable[QuartClient]]
parametrize = mark.parametrize


def random_string(length: int = 42) -> str:
    """Return a random printable string of given length."""
    return ''.join(choices(printable, k=42))


@fixture(name='app')
async def acanban() -> AsyncIterator[Acanban]:
    async with app.test_app() as test_app: yield test_app


@fixture
def client(app: Acanban) -> QuartClient:
    """Return a Quart test client."""
    return app.test_client()


@fixture
async def user(client: QuartClient) -> ClientFactory:
    """Return a coroutine which gives a logged client."""
    async def login(username: Optional[str]) -> QuartClient:
        """Return the logged client if username is not None."""
        if username is None: return client  # I'm sorry Nathaniel!
        form = {'username': username, 'password': username[::-1]}
        await client.post('/login', form=form)
        return client

    return login



# Authentication handler
# Copyright (C) 2020  Nguyễn Gia Phong
# Copyright (C) 2020  Ngô Ngọc Đức Huy
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from crypt import crypt
from hmac import compare_digest
from typing import Optional

from quart import (Blueprint, ResponseReturnValue, current_app,
                   redirect, render_template, request)
from quart_auth import AuthManager, AuthUser, login_user, logout_user
from rethinkdb import r
from rethinkdb.errors import ReqlNonExistenceError

from .db import RethinkObject

__all__ = ['Authenticator', 'blueprint']

ROLES = 'assistant', 'student', 'supervisor'
blueprint = Blueprint('auth', __name__)


class User(AuthUser, RethinkObject):
    """Acanban user.

    Properties needs to be awaited and ReqlNonExistenceError
    might be raised if the property or the user itself does not exist.

    Parameters
    ----------
    username : Optional[str]
        User's authentication identification.

    Attributes
    ----------
    username : str
        Pseudonym to refer to the user.
    password : str
        Hash of the user's password.
    name : str
        Real name of the user.
    email : str
        Email of the user.
    role : str
        Role of the user.
    """

    slots = 'username', 'password', 'name', 'email', 'role', 'projects'
    table = 'users'

    def __init__(self, username: Optional[str]) -> None:
        super().__init__(username)
        self.key = '' if username is None else username


class Authenticator(AuthManager):
    user_class = User


async def add_user(username: str, password: str,
                   name: str, email: str, role: str) -> None:
    """Try to add user to database."""
    # This never happens through browser, but via manual requests.
    if role not in ROLES: raise ValueError('unknown role')
    user = {'username': username, 'password': crypt(password),
            'name': name, 'email': email, 'role': role}
    async with current_app.db_pool.connection() as connection:
        user_with_mail = await r.table('users').filter(
            {'email': email}).count().run(connection)
        if user_with_mail > 0:
            raise ValueError('email taken')
        if (await r.table('users').insert(user).run(connection))['errors']:
            raise ValueError('username taken')


@blueprint.route('/register', methods=['GET', 'POST'])
async def register() -> ResponseReturnValue:
    """Handle the registration page."""
    if request.method == 'GET': return await render_template('register.html')
    info = await request.form
    try:
        await add_user(info['username'], info['password'],
                       info['name'], info['email'], info['role'])
    except ValueError as e:
        return await render_template('register.html', error=str(e))
    else:
        return redirect('/')


async def log_user(username: str, password: str) -> User:
    """Try to log the given user in."""
    user = User(username)
    try:
        digest = await user.password
    except ReqlNonExistenceError:
        raise ValueError('user does not exist')
    if not compare_digest(digest, crypt(password, digest)):
        raise ValueError('bad login')
    return user


@blueprint.route('/login', methods=['GET', 'POST'])
async def login() -> ResponseReturnValue:
    """Handle the login page."""
    if request.method == 'GET': return await render_template('login.html')
    info = await request.form
    username, password = info['username'], info['password']
    try:
        user = await log_user(username, password)
    except ValueError as e:
        return await render_template('login.html', error=str(e))
    else:
        remember = info.get('remember') is not None
        login_user(user, remember)
        return redirect('/')


@blueprint.route("/logout")
async def logout() -> ResponseReturnValue:
    """Handle logout."""
    logout_user()
    return redirect('/')



from typing import Any, AsyncContextManager

class Connection: ...

class TrioConnectionPool:
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    async def connection(self) -> AsyncContextManager[Connection]: ...
    async def close(self) -> None: ...



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options.
# For a full list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from os.path import join
from pathlib import Path

pwd = Path(__name__).parent

# Project information
project = 'Acanban'
copyright = '2020  Ngô Ngọc Đức Huy et al.'  # noqa
author = 'Ngô Ngọc Đức Huy et al.'

# Add any Sphinx extension module names here, as strings.
# They can be extensions coming with Sphinx (named 'sphinx.ext.*')
# or your custom ones.
extensions = ['sphinx.ext.extlinks', 'sphinx.ext.githubpages',
              'sphinxcontrib.plantuml', 'sphinxcontrib.inkscapeconverter']
extlinks = {'doi': ('https://doi.org/%s', 'doi: ')}
plantuml_output_format = 'svg_img'
plantuml_latex_output_format = 'pdf'

rst_epilog = """
.. Non-breaking space
.. |~| unicode:: 0xA0
   :trim:
"""

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# List of patterns, relative to source directory, that match
# files and directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# Options for HTML output
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets)
# here, relative to this directory.  They are copied after the builtin
# static files, so a file named "default.css" will overwrite the builtin
# "default.css".
html_static_path = []

# Options for LaTeX
latex_toplevel_sectioning = 'section'
latex_docclass = {'manual': 'article'}
latex_elements = {
    'papersize': 'a4paper', 'pointsize': '12pt',
    'babel': r'\usepackage[english,vietnamese]{babel}',
    'fontpkg': r'\usepackage{lmodern}',
    'fncychap': '',  # disable fancy chapter heading
    'preamble': r'\usepackage{acanban}',
    'extraclassoptions': 'onepage',
    'extrapackages': r'\usepackage{booktabs}',  # for title-page
    'geometry': '',  # fallback to LaTeX default margins
    'maketitle': (pwd/'latex'/'title-page.tex').read_text(),
    'tableofcontents': r'\selectlanguage{english}\sphinxtableofcontents',
    'printindex': '', 'sphinxsetup': r'HeaderFamily=\bfseries'}
latex_appendices = [join('appendix', 'acknowledgement'),
                    join('appendix', 'glossary')]
latex_additional_files = [join('latex', 'acanban.sty'),
                          join('latex', 'usth-logo.pdf')]

numfig = True
numfig_format = {'figure': 'Figure %s', 'table': 'Table %s'}



# Test task-specific routes
# Copyright (C) 2021  Nguyễn Gia Phong
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from http import HTTPStatus as Status
from typing import Optional

from conftest import ClientFactory, parametrize, random_string
from pytest import param

# Some members: evelynd (student), ronanf (supervisor)
# Some nonmembers: adaml (student), silasl (assistant)
BASE_ROUTE = '/p/7fee498b-98ac-4173-b600-e94618f9ea1f/tasks'


@parametrize(('username', 'status_code'),
             (param(None, Status.UNAUTHORIZED, id='guest'),
              param('silasl', Status.FORBIDDEN, id='assistant'),
              param('adaml', Status.FORBIDDEN, id='nonmember'),
              param('ronanf', Status.OK, id='member')))
async def test_create_get(username: Optional[str], status_code: int,
                          user: ClientFactory) -> None:
    """Test task creation get permission."""
    client = await user(username)
    response = await client.get(f'{BASE_ROUTE}/create')
    assert response.status_code == status_code


@parametrize(('username', 'assignee', 'status_code'),
             (param(None, 'evelynd', Status.UNAUTHORIZED, id='guest try'),
              param('silasl', 'evelynd', Status.FORBIDDEN, id='assistant try'),
              param('adaml', 'evelynd', Status.FORBIDDEN, id='nonmember try'),
              param('ronanf', 'nonexist', Status.BAD_REQUEST, id='assign err'),
              param('ronanf', 'evelynd', Status.FOUND, id='success')))
async def test_create_post(username: Optional[str], assignee: str,
                           status_code: int, user: ClientFactory) -> None:
    """Test task creation post permission."""
    client = await user(username)
    response = await client.post(f'{BASE_ROUTE}/create', form=dict(
        name=random_string(), description=random_string(),
        assignee=assignee, deadline='2038-01-19'))
    assert response.status_code == status_code


@parametrize(('username', 'status_code'),
             (param(None, Status.UNAUTHORIZED, id='guest'),
              param('silasl', Status.FORBIDDEN, id='assistant'),
              param('adaml', Status.FORBIDDEN, id='nonmember'),
              param('ronanf', Status.OK, id='member')))
async def test_get(username: Optional[str], status_code: int,
                   user: ClientFactory) -> None:
    """Test task access permission."""
    client = await user(username)
    response = await client.get(f'{BASE_ROUTE}/0/')
    assert response.status_code == status_code


@parametrize(('username', 'status_code'),
             (param(None, Status.UNAUTHORIZED, id='guest'),
              param('silasl', Status.FORBIDDEN, id='assistant'),
              param('adaml', Status.FORBIDDEN, id='nonmember'),
              param('ronanf', Status.FOUND, id='member')))
@parametrize('direction', ('dec', 'inc'))
async def test_move(username: Optional[str], status_code: int,
                    user: ClientFactory, direction: str) -> None:
    """Test moving task across columns."""
    client = await user(username)
    response = await client.get(f'{BASE_ROUTE}/0/{direction}')
    assert response.status_code == status_code


@parametrize(('username', 'comment', 'status_code'),
             (param(None, 1609023842.0, Status.UNAUTHORIZED, id='guest'),
              param('silasl', 1609023842.0, Status.FORBIDDEN, id='assistant'),
              param('adaml', 1609023842.0, Status.FORBIDDEN, id='nonmember'),
              param('ronanf', 1609023842.0, Status.FOUND, id='member'),
              param('ronanf', 420.69, Status.NOT_FOUND, id='unknown comment')))
async def test_reply(username: Optional[str], comment: float,
                     status_code: int, user: ClientFactory) -> None:
    """Test replying to a comment."""
    client = await user(username)
    response = await client.post(f'{BASE_ROUTE}/0/reply/{comment}',
                                 form={'comment': 'Alright.'})
    assert response.status_code == status_code



# Test authentication handling
# Copyright (C) 2020  Nguyễn Gia Phong
# Copyright (C) 2020  Ngô Ngọc Đức Huy
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from http import HTTPStatus

from pytest import mark
from quart.testing import QuartClient


@mark.parametrize(('username', 'role', 'code'),
                  (('kiddo', 'pupil', 200),
                   ('silasl', 'supervisor', 200),
                   ('huyngo', 'student', 302)))
async def test_register(username: str, role: str, code: int,
                        client: QuartClient) -> None:
    """Test successful and failed registration."""
    response = await client.post('/register', form=dict(
        username=username, password='baz',
        name='Fô Bả', email='f@o.o', role=role))
    # Successful registration redirects.
    assert response.status_code == code


async def test_register_existing_email(client: QuartClient) -> None:
    """Test failed registration where the email is taken."""
    response = await client.post('/register', form=dict(
        username='notronanf', password='some-password',
        name='Ronan Franklin', email='ronanf@example.edu', role='supervisor'))
    assert response.status_code == HTTPStatus.OK


@mark.parametrize(('username', 'password', 'code'),
                  (('huyngo', 'baz', 302), ('silasl', 'lsalis', 302),
                   ('opheliad', 'wrong', 200), ('ne-existe', 'pas', 200)))
async def test_login(username: str, password: str, code: int,
                     client: QuartClient) -> None:
    """Test successful and failed login."""
    response = await client.post(
        '/login', form=dict(username=username, password=password))
    # Successful login redirects.
    assert response.status_code == code



from typing import Any, Dict, Union

from .trio_net.net_trio import Connection

class RqlQuery:
    def __getitem__(self, index: Any) -> RqlQuery: ...

    def count(self) -> RqlQuery: ...
    def filter(self, predicate: dict) -> RqlQuery: ...
    def pluck(self, *selectors: Union[str, Dict[str, Any]]) -> RqlQuery: ...
    def order_by(self, key: str) -> RqlQuery: ...
    def update(self, obj: dict) -> RqlQuery: ...

    # Loaded JSON or TrioCursor of loaded JSON,
    # static typing has no power here.
    async def run(self, c: Connection) -> Any: ...

class Table(RqlQuery):
    def get(self, key: str) -> RqlQuery: ...
    def get_all(self, *keys: str, index: str = 'id') -> RqlQuery: ...
    def insert(self, *objects: dict) -> RqlQuery: ...



from typing import Optional

def user_config_dir(appname: Optional[str] = None,
                    appauthor: Optional[str] = None,
                    version: Optional[str] = None,
                    roaming: Optional[bool] = False) -> str: ...
def site_config_dir(appname: Optional[str] = None,
                    appauthor: Optional[str] = None,
                    version: Optional[str] = None,
                    multipath: Optional[bool] = False) -> str: ...



# Initialization code
# Copyright (C) 2020-2021  Nguyễn Gia Phong
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from contextlib import asynccontextmanager
from datetime import datetime, timedelta, timezone as tz
from secrets import token_urlsafe
from time import timezone
from typing import Any, AsyncIterator
from urllib.parse import urlsplit

from bleach import clean, linkify
from bleach_allowlist import markdown_attrs, markdown_tags
from httpx import AsyncClient
from humanize import naturalsize, naturaltime
from markdown import markdown
from quart import ResponseReturnValue, current_app, render_template
from quart_auth import current_user
from quart_trio import QuartTrio
from rethinkdb import r
from rethinkdb.trio_net.net_trio import TrioConnectionPool
from trio import open_nursery

from .auth import Authenticator, blueprint as auth
from .config import IPFS_DEFAULT, RETHINKDB_DEFAULT
from .ipfs import blueprint as ipfs
from .project import PREVIEW_FIELDS, blueprint as project
from .task import blueprint as task
from .user import blueprint as user

__all__ = ['app']
__doc__ = 'Academic Kanban'
__version__ = '0.1.0'

# Nope, the negative operator is not a typo, see also tzset(3posix).
TZ = tz(timedelta(seconds=-timezone))


class Acanban(QuartTrio):
    """Acanban web application."""

    auth_manager: Authenticator
    db_config = RETHINKDB_DEFAULT
    db_pool: TrioConnectionPool
    ipfs_config = IPFS_DEFAULT
    ipfs_gateway: AsyncClient
    ipfs_api: AsyncClient

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.secret_key = token_urlsafe(16)
        Authenticator().init_app(self)

    @asynccontextmanager
    async def test_app(self) -> AsyncIterator[Acanban]:
        """Return a context manager startup and shutdown the app."""
        async with open_nursery() as nursery:
            self.nursery = nursery
            await self.startup()
            try:
                yield self
            finally:
                await self.shutdown()
        self.nursery = None


app = Acanban(__name__)
app.register_blueprint(auth)
app.register_blueprint(ipfs)
app.register_blueprint(user)
app.register_blueprint(project)
app.register_blueprint(task)
app.add_template_filter(naturalsize)


@app.template_filter('markdown')
def as_markdown(text: str) -> str:
    html = clean(markdown(linkify(text)), markdown_tags, markdown_attrs)
    return html


@app.template_filter('naturaltime')
def fuzzytime(dt: datetime) -> str:
    """Convert datetime into fuzzy time in HTML with precise hovertext."""
    rounded = dt.replace(microsecond=0)
    fuzzy = naturaltime(rounded.replace(tzinfo=None), when=datetime.utcnow())
    exact = rounded.astimezone(TZ)
    return f"<span title='{exact}'>{fuzzy}</span>"


@app.before_serving
async def create_db_pool() -> None:
    """Create RethinkDB connection pool."""
    r.set_loop_type('trio')
    app.db_pool = r.ConnectionPool(nursery=app.nursery, **app.db_config)


@app.after_serving
async def close_db_pool() -> None:
    """Close RethinkDB connection pool."""
    await app.db_pool.close()


@app.before_serving
async def open_ipfs_clients() -> None:
    """Open HTTPX client for IPFS gateway and HTTP API."""
    app.ipfs_gateway = AsyncClient(base_url=app.ipfs_config['gateway']['base'])
    api_url = app.ipfs_config['api']['base']
    app.ipfs_api = AsyncClient(base_url=api_url, headers={
        'Accept': 'application/json', 'Host': urlsplit(api_url)[1],
        'User-Agent': f'acanban/{__version__}'})


@app.after_serving
async def close_ipfs_clients() -> None:
    """Close HTTPX client for IPFS gateway and HTTP API."""
    await app.ipfs_gateway.aclose()
    await app.ipfs_api.aclose()


@app.route('/')
async def index() -> ResponseReturnValue:
    """Return the index page."""
    if await current_user.is_authenticated:
        username = current_user.key
        my_projects = r.table('projects').get_all(username, index='members')
        async with current_app.db_pool.connection() as connection:
            projects = await my_projects.pluck(*PREVIEW_FIELDS).run(connection)
        return await render_template('myproject.html', projects=projects)
    else:
        return await render_template('index.html')



from typing import Any, Optional

class ReqlError(Exception):
    def __init__(self, message: str, term: Optional[Any],
                 frames: Optional[Any]) -> None: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...

class ReqlNonExistenceError(ReqlError): ...



# Test project pages
# Copyright (C) 2020-2021  Nguyễn Gia Phong
# Copyright (C) 2020  Ngô Xuân Minh
# Copyright (C) 2021  Ngô Ngọc Đức Huy
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from http import HTTPStatus as Status
from random import uniform
from typing import Optional

from conftest import ClientFactory, parametrize, random_string
from pytest import param, raises
from quart import Blueprint

from acanban.project import add_artifact_tab

# Some members: adaml (student), oliviak (supervisor)
# Some nonmembers: ronanf (supervisor), silasl (assistant)
PROJECT = 'be7a04b8-650f-41dc-912b-10d225baff29'


@parametrize('tab', ('info', 'edit', 'members', 'tasks', 'report', 'slides'))
async def test_nonexist(tab: str, user: ClientFactory) -> None:
    """Test accessing tabs on a nonexistent project."""
    adaml = await user('adaml')
    response = await adaml.get(f'/p/this-project-does-not-exist/{tab}')
    assert response.status_code == Status.NOT_FOUND


@parametrize(('username', 'status_code'),
             (param(None, Status.UNAUTHORIZED, id='guest'),
              param('silasl', Status.FORBIDDEN, id='assistant'),
              param('ronanf', Status.OK, id='supervisor'),
              param('adaml', Status.OK, id='student')))
async def test_create_get(username: Optional[str], status_code: int,
                          user: ClientFactory) -> None:
    """Test project creation page access permission."""
    client = await user(username)
    response = await client.get('/p/create')
    assert response.status_code == status_code


@parametrize(('username', 'status_code'),
             (param(None, Status.UNAUTHORIZED, id='guest'),
              param('silasl', Status.FORBIDDEN, id='assistant'),
              param('ronanf', Status.FOUND, id='supervisor'),
              param('adaml', Status.FOUND, id='student')))
async def test_create_post(username: Optional[str], status_code: int,
                           user: ClientFactory) -> None:
    """Test project creation permission."""
    client = await user(username)
    info = {'name': random_string(), 'deadline': '2020-02-02',
            'description': random_string()}
    response = await client.post('/p/create', form=info)
    assert response.status_code == status_code


@parametrize(('username', 'status_code'),
             (param(None, Status.UNAUTHORIZED, id='guest'),
              param('silasl', Status.FORBIDDEN, id='assistant'),
              param('ronanf', Status.FORBIDDEN, id='nonmember'),
              param('adaml', Status.OK, id='member')))
@parametrize('tab', ('info', 'members', 'edit', 'tasks', 'report', 'slides'))
async def test_get(username: Optional[str], status_code: int,
                   tab: str, user: ClientFactory) -> None:
    """Test project tabs access permission."""
    client = await user(username)
    response = await client.get(f'/p/{PROJECT}/{tab}')
    assert response.status_code == status_code


async def test_info_redirect(user: ClientFactory) -> None:
    """Test redirect from project's root URI."""
    adaml = await user('adaml')
    response = await adaml.get(f'/p/{PROJECT}/')
    assert response.status_code == Status.FOUND


@parametrize(('username', 'status_code'),
             (param(None, Status.UNAUTHORIZED, id='guest'),
              param('silasl', Status.FORBIDDEN, id='assistant'),
              param('ronanf', Status.FORBIDDEN, id='nonmember'),
              param('adaml', Status.FOUND, id='member')))
async def test_edit_post(username: Optional[str], status_code: int,
                         user: ClientFactory) -> None:
    """Test project edit permission."""
    client = await user(username)
    info = {'name': random_string(), 'deadline': '2021-12-02',
            'description': random_string()}
    response = await client.post(f'/p/{PROJECT}/edit', form=info)
    assert response.status_code == status_code


@parametrize(('username', 'invited', 'status_code'), (
    param(None, 'ronanf', Status.UNAUTHORIZED, id='guest invites'),
    param('silasl', 'ronanf', Status.FORBIDDEN, id='assistant invites'),
    param('ronanf', 'ronanf', Status.FORBIDDEN, id='nonmember invites'),
    param('adaml', 'nonexist', Status.OK, id='invite nonexist'),
    param('adaml', 'oliviak', Status.OK, id='invite member'),
    param('adaml', 'silasl', Status.OK, id='invite assistant'),
    param('adaml', 'ronanf', Status.FOUND, id='invite success')))
async def test_member_invite(username: Optional[str], invited: str,
                             status_code: int, user: ClientFactory) -> None:
    """Test member invite permission."""
    client = await user(username)
    new_user = {'new-user': invited}
    response = await client.post(f'/p/{PROJECT}/invite', form=new_user)
    assert response.status_code == status_code


@parametrize(('username', 'status_code'),
             (param(None, Status.UNAUTHORIZED, id='guest'),
              param('silasl', Status.FORBIDDEN, id='nonstudent nonmember'),
              param('lucyl', Status.FORBIDDEN, id='student nonmember'),
              param('oliviak', Status.FORBIDDEN, id='nonstudent member'),
              param('adaml', Status.FOUND, id='student member')))
@parametrize('tab', ('report', 'slides'))
async def test_artifact_upload(username: Optional[str], status_code: int,
                               tab: str, user: ClientFactory) -> None:
    """Test report and slides upload."""
    response = await (await user(username)).post(
        f'/p/{PROJECT}/{tab}/upload',
        headers={
            'Content-Length': 199,
            'Content-Type': (
                'multipart/form-data;'
                ' boundary=------------------------e70696c7f3938bcf')},
        data=(
            b'--------------------------e70696c7f3938bcf\r\n'
            b'Content-Disposition: form-data; name="file"; filename="foo"\r\n'
            b'Content-Type: application/octet-stream\r\n\r\n'
            b'bar\n\r\n--------------------------e70696c7f3938bcf--\r\n'))
    assert response.status_code == status_code


@parametrize(('username', 'status_code'),
             (param(None, Status.UNAUTHORIZED, id='guest'),
              param('silasl', Status.FORBIDDEN, id='nonstudent nonmember'),
              param('lucyl', Status.FORBIDDEN, id='student nonmember'),
              param('adaml', Status.FORBIDDEN, id='student member'),
              param('oliviak', Status.FOUND, id='nonstudent member')))
@parametrize('tab', ('report', 'slides'))
async def test_artifact_eval(username: Optional[str], status_code: int,
                             tab: str, user: ClientFactory) -> None:
    """Test report and presentation evaluation."""
    client = await user(username)
    evaluation = {'grade': uniform(0, 20), 'comment': random_string()}
    response = await client.post(f'/p/{PROJECT}/report/eval', form=evaluation)
    assert response.status_code == status_code


def test_add_invalid_artifact_tab() -> None:
    """Test adding an artifact tab that is neither report nor slides."""
    with raises(ValueError): add_artifact_tab(Blueprint('foo', 'bar'), 'baz')



# Test basic routing
# Copyright (C) 2020  Nguyễn Gia Phong
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from conftest import ClientFactory, parametrize
from quart.testing import QuartClient


@parametrize(('uri', 'status_code'), (
    ('/', 200), ('/p/', 200),
    ('/register', 200), ('/login', 200), ('/logout', 302),
    ('/ipfs/QmPTWcNNxPdVwqQ939rsxJQiiditfw9YNyLMs9QZJM4gCU', 200),
    ('/ipfs/QmRW3V9znzFW9M5FYbitSEvd5dQrPWGvPvgQD6LM22Tv8D', 301),
    ('/ipns/ipfs.io', 301), ('/ipfs/ipfs.io', 404)))
async def test_status(uri: str, status_code: int, client: QuartClient) -> None:
    """Test the status of basic routes."""
    response = await client.get(uri)
    assert response.status_code == status_code
    await response.get_data()


async def test_myproject_status(user: ClientFactory) -> None:
    """Test the status of user dashboard."""
    client = await user('adaml')
    response = await client.get('/')
    assert response.status_code == 200



# Test RethinkDB wrapper
# Copyright (C) 2020  Ngô Ngọc Đức Huy
# Copyright (C) 2020  Nguyễn Gia Phong
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from crypt import crypt
from hmac import compare_digest

from pytest import raises

from acanban import Acanban
from acanban.auth import User


async def test_object_get(app: Acanban) -> None:
    """Test object get attributes."""
    async with app.app_context():
        user = User('silasl')
        digest = await user.password
        assert compare_digest(digest, crypt('lsalis', digest))
        assert await user.email == 'silasl@example.edu'
        assert await user.name == 'Silas Lehtonen'
        assert await user.role == 'assistant'
        with raises(AttributeError): await user.kink


async def test_object_set(app: Acanban) -> None:
    """Test object set attributes."""
    async with app.app_context():
        user, monty = User('silasl'), 'Monty Python'
        name = await user.name
        with raises(AttributeError): user.name = monty

        await user.update(name=monty)
        assert await user.name == monty
        await user.update(name=name)

        with raises(ValueError): await user.update(kink='DBMS')



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['acanban']

package_data = \
{'': ['*'], 'acanban': ['static/*', 'templates/*']}

package_dir = \
{'': 'src'}

install_requires = \
['appdirs',
 'bleach',
 'bleach_allowlist',
 'httpx',
 'humanize',
 'hypercorn',
 'markdown',
 'py-cid',
 'quart-auth',
 'quart-trio-twice',
 'rethinkdb >=2.4.8']

setup(name='acanban',
      version='0.1.0',
      description='Academic Kanban',
      author='Ngô Ngọc Đức Huy',
      author_email='duchuy29092000@gmail.com',
      url='https://Huy-Ngo.github.io/acanban',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      python_requires='>=3.7',
     )



# Project pages
# Copyright (C) 2020-2021  Nguyễn Gia Phong
# Copyright (C) 2021  Ngô Ngọc Đức Huy
# Copyright (C) 2021  Ngô Xuân Minh
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from functools import partial
from itertools import tee
from operator import itemgetter
from typing import Any, Dict, Sequence, Union

from quart import (Blueprint, ResponseReturnValue, current_app,
                   redirect, render_template, request)
from quart.exceptions import Forbidden, NotFound
from quart_auth import current_user, login_required
from rethinkdb import r
from rethinkdb.errors import ReqlNonExistenceError

from .ipfs import add as ipfs_add

__all__ = ['PREVIEW_FIELDS', 'blueprint']

INFO_FIELDS = 'id', 'name', 'created_on', 'deadline', 'description'
MEMBERS_FIELDS = 'id', 'name', 'supervisors', 'students'
PREVIEW_FIELDS = ('id', 'name', 'created_on', 'deadline',
                  'supervisors', 'students', 'description')
TASKS_FIELDS = 'id', 'name', {'tasks': ('name', 'created_on', 'status')}

PluckKey = Union[str, Dict[str, Any]]
blueprint = Blueprint('project', __name__, url_prefix='/p')


@blueprint.route('/create', methods=['GET', 'POST'])
@login_required
async def create() -> ResponseReturnValue:
    role = await current_user.role
    if role == 'assistant':
        raise Forbidden
    if request.method == 'GET':
        return await render_template('project-create.html')
    project = await request.form
    project = {
        'name': project['name'], 'description': project['description'],
        'students': [], 'supervisors': [], f'{role}s': [current_user.key],
        'created_on': r.now(),  # RethinkDB.now defaults to UTC
        'deadline': r.iso8601(project['deadline'], default_timezone='+00:00'),
        'tasks': [], 'report': {'revisions': []}, 'slides': {'revisions': []}}
    async with current_app.db_pool.connection() as connection:
        response = await r.table('projects').insert(project).run(connection)
        uuid = response['generated_keys'][0]
    return redirect(f'/p/{uuid}')


@blueprint.route('/')
async def list_projects() -> ResponseReturnValue:
    """Return a page listing all projects."""
    project_list = r.table('projects').pluck(*PREVIEW_FIELDS)
    async with current_app.db_pool.connection() as connection:
        projects = await project_list.run(connection)
    return await render_template('project-list.html', projects=projects)


@blueprint.route('/<uuid>/')
@login_required
async def redirect_info(uuid: str) -> ResponseReturnValue:
    """Redirect to the project info page."""
    return redirect(f'/p/{uuid}/info')


async def pluck(uuid: str, fields: Sequence[PluckKey] = ()) -> Dict[str, Any]:
    """Pluck the given fields from the project, with permission check."""
    async with current_app.db_pool.connection() as connection:
        try:
            project = await r.table('projects').get(uuid).pluck(
                'students', 'supervisors', *fields).run(connection)
        except ReqlNonExistenceError:
            raise NotFound

    members = project['students'] + project['supervisors']
    if current_user.key not in members: raise Forbidden
    return {key: value for key, value in project.items()
            if any(key == field or key in field for field in fields)}


@blueprint.route('/<uuid>/info')
@login_required
async def show_info(uuid: str) -> ResponseReturnValue:
    """Return the page containing the projects' basic infomation."""
    project = await pluck(uuid, INFO_FIELDS)
    return await render_template('project-info.html', project=project)


@blueprint.route('/<uuid>/edit', methods=['GET', 'POST'])
@login_required
async def edit(uuid: str) -> ResponseReturnValue:
    """Return the page for editting the projects' basic infomation."""
    if request.method == 'GET':
        project = await pluck(uuid, INFO_FIELDS)
        return await render_template('project-edit.html', project=project)

    await pluck(uuid)  # check project's existence and permission
    raw = await request.form
    updated = dict(
        name=raw['name'], description=raw['description'],
        deadline=r.iso8601(raw['deadline'], default_timezone='+00:00'))
    async with current_app.db_pool.connection() as conn:
        await r.table('projects').get(uuid).update(updated).run(conn)
    return redirect(f'/p/{uuid}/info')


@blueprint.route('/<uuid>/members')
@login_required
async def list_members(uuid: str) -> ResponseReturnValue:
    """Return the page listing the member of a project."""
    project = await pluck(uuid, MEMBERS_FIELDS)
    return await render_template('project-members.html', project=project)


@blueprint.route('/<uuid>/invite', methods=['POST'])
@login_required
async def invite_member(uuid: str) -> ResponseReturnValue:
    """Add a member to the project."""
    username = (await request.form)['new-user']
    project = await pluck(uuid, MEMBERS_FIELDS)
    render = partial(render_template, 'project-members.html', project=project)

    async with current_app.db_pool.connection() as conn:
        try:
            role = await r.table('users').get(username)['role'].run(conn)
        except ReqlNonExistenceError:
            return await render(error=f'{username} has not registered')
    if role == 'assistant':
        return await render(error=f'{username} is an assistant')
    mem_field = f'{role}s'
    if username in project[mem_field]:
        return await render(error=f'{username} is already a member')

    async with current_app.db_pool.connection() as conn:
        await r.table('projects').get(uuid).update(
            {mem_field: r.row[mem_field].append(username)}).run(conn)
        await r.table('users').get(username).update(
            {'projects': r.row['projects'].append(uuid)}).run(conn)
    return redirect(request.referrer)


@blueprint.route('/<uuid>/tasks')
@login_required
async def list_tasks(uuid: str) -> ResponseReturnValue:
    """Return the page listing tasks in a Kanban board."""
    project = await pluck(uuid, TASKS_FIELDS)
    tasks = sorted(project['tasks'], key=itemgetter('created_on'))
    columns = tee(enumerate(tasks), 3)
    todo, doing, done = ([{'index': index, **task} for index, task in column
                          if status == task['status']]
                         for status, column in enumerate(columns))
    return await render_template('project-tasks.html', project=project,
                                 todo=todo, doing=doing, done=done)


async def show_artifacts(tab: str, uuid: str) -> ResponseReturnValue:
    """Return the tab with artifacts upload and evaluation."""
    await pluck(uuid)  # check project's existence and permission
    project = await pluck(uuid, ('id', 'name', tab))
    query = r.table('files').get_all(*project[tab]['revisions'])
    async with current_app.db_pool.connection() as conn:
        revisions = [file async for file in await query.run(conn)]
    revisions.sort(key=itemgetter('time'), reverse=True)
    variables = {'project': project, tab: project[tab], 'revisions': revisions,
                 'for_student': (await current_user.role)=='student'}
    return await render_template(f'project-{tab}.html', **variables)


async def upload_artifact(tab: str, uuid: str) -> ResponseReturnValue:
    """Handle file upload to the corresponding tab."""
    await pluck(uuid)  # check project's existence and permission
    if await current_user.role != 'student': raise Forbidden
    action = r.row[tab]['revisions'].append(await ipfs_add())
    async with current_app.db_pool.connection() as conn:
        await r.table('projects').get(uuid).update(
            {tab: {'revisions': action}}).run(conn)
    return redirect(request.referrer)


async def eval_artifact(tab: str, uuid: str) -> ResponseReturnValue:
    """Handle evaluation of the work in the corresponding tab."""
    await pluck(uuid)  # check project's existence and permission
    if await current_user.role == 'student': raise Forbidden
    form = await request.form
    updated = {'grade': float(form['grade']), 'comment': form['comment']}
    query = r.table('projects').get(uuid).update({tab: updated})
    async with current_app.db_pool.connection() as conn: await query.run(conn)
    return redirect(request.referrer)


def add_artifact_tab(blueprint: Blueprint, tab: str) -> None:
    """Add tab of the given name to the blueprint.

    Raise ValueError if tab is neither report nor slides.
    """
    if tab not in ('report', 'slides'):
        raise ValueError(f'tab is neither report nor slides: {tab}')
    blueprint.add_url_rule(
        f'/<uuid>/{tab}', methods=['GET'], endpoint=f'{tab}_view',
        view_func=login_required(partial(show_artifacts, tab)))
    blueprint.add_url_rule(
        f'/<uuid>/{tab}/upload', methods=['POST'], endpoint=f'{tab}_upload',
        view_func=login_required(partial(upload_artifact, tab)))
    blueprint.add_url_rule(
        f'/<uuid>/{tab}/eval', methods=['POST'], endpoint=f'{tab}_eval',
        view_func=login_required(partial(eval_artifact, tab)))


add_artifact_tab(blueprint, 'report')
add_artifact_tab(blueprint, 'slides')



from datetime import datetime
from typing import Optional, Union

def naturalsize(value: Union[float, str], binary: bool = False,
                gnu: bool = False, format: str = '%.1f') -> str: ...
def naturaltime(value: Union[datetime, int], future: bool = False,
                months: bool = True, minimum_unit: str = 'seconds',
                when: Optional[datetime] = None) -> str: ...
